@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: address.c,v 1.33 1999/10/16 06:08:39 motonori Exp $";
+static char *_id_ = "$Id: address.c,v 1.41 2002/01/20 14:19:39 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -240,8 +240,9 @@ u_char *addr;
 	}
 	bcopy(addr, iap->address, len);
 #if INET6
-	/* put IPv6 address at first */
-	if (inet_domain == AF_INET6 || tailp == NULL)
+	if (((cnf.inetdom & SMTP_V6_FIRST) && inet_domain == AF_INET6)
+	 || ((cnf.inetdom & SMTP_V4_FIRST) && inet_domain == AF_INET)
+	 || tailp == NULL)
 	{
 		iap->next = hostp->addr;
 		hostp->current = hostp->addr = iap;
@@ -356,7 +357,7 @@ char **err, **addr, **dom;
 		if (isspace(*str))
 			break;
 		if (isalnum(*str)
-		 || strchr("!.#$%&*-+/=?_~{}'`|", *str) != NULL)
+		 || strchr("!.#$%&^*-+/=?_~{}'`|", *str) != NULL)
 		{
 			if (gotpure)
 				str++;
@@ -498,6 +499,7 @@ char *addr, *domain, *notify, *orcpt;
 		rcpt->domain->name = newstr(domain);
 		if (rcpt->domain->name == NULL)
 			return -1;
+		rcpt->domain->rcpts = hash_domain_rcpts_lookup(domain);
 		rcpt->domain->rcpt_top = rcpt;
 		rcpt->domain->rcpt_tail = rcpt;
 		rcpt->domain->next = domain_list;
@@ -538,14 +540,24 @@ char *addr, *domain, *notify, *orcpt;
 	return 0;
 }
 
+void
+dnsstatus(rcpt, rcode, response)
+struct recipient *rcpt;
+int rcode;
+char *response;
+{
+	log(LOG_INFO, "(%d+%d+%d+%d/%d) relay=%s to=%s proto=%s delay=%d code=%d (%s)",            
+		sti.nsent, sti.ndeferred, sti.nnsfailed,
+		sti.nsmtpfailed, sti.nrcpt, "unknown",
+		rcpt->address, "unknown", 0, rcode,   
+		(response==NULL)?"?":response);   
+}
+
 void
 finalstatus()
 {
 	struct recipient *rcpt;
 	struct domain *dom;
-#if 0
-	struct mx *mxp;
-#endif
 
 	for (rcpt = env.rcpt_list; rcpt != NULL; rcpt = rcpt->next)
 	{
@@ -557,6 +569,7 @@ finalstatus()
 			rcpt->stat = RCPT_DONE;
 			rcpt->result = SMTP_TEMPFAIL(51);
 			rcpt->response = "Unknown system error (domain)";
+			dnsstatus(rcpt, rcpt->result, rcpt->response);
 			continue;
 		}
 		switch (dom->stat)
@@ -567,24 +580,28 @@ finalstatus()
 			rcpt->stat = RCPT_DONE;
 			rcpt->result = SMTP_TEMPFAIL(52);
 			rcpt->response = dom->response;
+			dnsstatus(rcpt, rcpt->result, rcpt->response);
 			continue;
 		    case EX_NOHOST:
 			sti.nnsfailed++;
 			rcpt->stat = RCPT_DONE;
 			rcpt->result = SMTP_ERROR(50);
 			rcpt->response = dom->response;
+			dnsstatus(rcpt, rcpt->result, rcpt->response);
 			continue;
 		    case EX_CONFIG:
 			sti.nnsfailed++;
 			rcpt->stat = RCPT_DONE;
 			rcpt->result = SMTP_ERROR(54);
 			rcpt->response = dom->response;
+			dnsstatus(rcpt, rcpt->result, rcpt->response);
 			continue;
 		    case EX_OSERR:
 		    default:
 			rcpt->stat = RCPT_DONE;
 			rcpt->result = SMTP_TEMPFAIL(51);
 			rcpt->response = "Unknown system error (domain)";
+			dnsstatus(rcpt, rcpt->result, rcpt->response);
 			continue;
 		}
 		if (dom->firstmx == NULL)
@@ -592,41 +609,8 @@ finalstatus()
 			rcpt->stat = RCPT_DONE;
 			rcpt->result = SMTP_TEMPFAIL(51);
 			rcpt->response = "Unknown system error (MX)";
+			dnsstatus(rcpt, rcpt->result, rcpt->response);
 			continue;
 		}
-#if 0
-		for (mxp = dom->firstmx; mxp != NULL; mxp = mxp->next)
-		{
-			switch (mxp->stat)
-			{
-			    case EX_OK:
-				break;
-			    case EX_TEMPFAIL:
-				rcpt->stat = RCPT_DONE;
-				rcpt->result = SMTP_TEMPFAIL(52);
-				rcpt->response = mxp->response;
-				goto cont;
-			    case EX_NOHOST:
-				sti.nnsfailed++;
-				rcpt->stat = RCPT_DONE;
-				rcpt->result = SMTP_ERROR(50);
-				rcpt->response = mxp->response;
-				goto cont;
-			    case EX_CONFIG:
-				sti.nnsfailed++;
-				rcpt->stat = RCPT_DONE;
-				rcpt->result = SMTP_ERROR(54);
-				rcpt->response = mxp->response;
-				goto cont;
-			    case EX_OSERR:
-			    default:
-				rcpt->stat = RCPT_DONE;
-				rcpt->result = SMTP_TEMPFAIL(51);
-				rcpt->response = "Unknown error on MX";
-				goto cont;
-			}
-		}
-  cont:;
-#endif
 	}
 }
