@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,7 +33,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: common.h,v 1.55 1999/03/02 21:59:57 motonori Exp $
+ * $Id: common.h,v 1.68 2002/01/20 14:20:42 motonori Exp $
  */
 
 # include "compat.h"
@@ -56,10 +56,10 @@
 # include <ctype.h>
 # include <sys/param.h>
 # include <sys/resource.h>
+# include <sys/socket.h>
 # include <netinet/in.h>
 # include <arpa/inet.h>
 # include <arpa/nameser.h>
-# include <sys/socket.h>
 # include <sys/stat.h>
 # include <setjmp.h>
 # include <resolv.h>
@@ -104,6 +104,9 @@
 #ifndef EX_CONFIG
 # define EX_CONFIG	78		/* sysexit.h */
 #endif
+#ifndef T_AAAA
+# define T_AAAA		28		/* IP6 Address */
+#endif
 
 #ifndef MAXHOSTNAMELEN
 # define MAXHOSTNAMELEN	256		/* sys/param.h */
@@ -158,9 +161,7 @@
 # define RR_MX			0x02
 # define RR_CNAME		0x04
 # define RR_A			0x08
-#ifdef INET6
 # define RR_AAAA		0x10
-#endif
 # define RR_ALL			0x1f
 
 # define MALLOC(x)		((sti.mem += (x)), malloc(x))
@@ -237,6 +238,7 @@ struct mx {
 	struct mx		*next;		/* next MX */
 	struct mx		*mx_ref;	/* MX reference list */
 	struct domain		*domain;	/* reverse link to domain */
+	long			opt;		/* option in map file */
 };
 
 struct host_map {
@@ -246,6 +248,8 @@ struct host_map {
 	struct host_map		*hash;		/* chain in hash table */
 };
 
+# define MAPOPT_TRYALLADDR	0001		/* try all addr even if TFAIL */
+
 struct domain {
 	char			*name;		/* domain part of mail addr */
 	int			stat;		/* error status for MX query */
@@ -259,6 +263,13 @@ struct domain {
 	struct domain		*workchain;	/* domain chain with job */
 	struct domain		*hash;		/* domain chain in hash table */
 	char			**map_arg;	/* way of host lookup */
+	int			rcpts;		/* max recipient count */
+};
+
+struct domain_rcpts {
+	char			*name;		/* destination domain */
+	int			rcpts;		/* count of recipients */
+	struct domain_rcpts	*hash;		/* chain in hash table */
 };
 
 # define RCPT_NOOP	0
@@ -339,6 +350,7 @@ struct statinfo {
 	int			nsmtpfailed;	/* total recps w/ err. (SMTP) */
 	int			minsock;	/* min# of used sockets */
 	int			maxsock;	/* max# of used sockets */
+	int			maxcquery;	/* max# of concurrent queries */
 	int			dns_inb;	/* total incoming DNS traffic */
 	int			dns_outb;	/* total outgoing DNS traffic */
 	int			inbounds;	/* total incoming traffic */
@@ -369,17 +381,21 @@ struct statinfo {
 #ifdef INET6			/* flags */
 # define SMTP_V4		0x1
 # define SMTP_V6		0x2
+# define SMTP_V4_FIRST		0x4
+# define SMTP_V6_FIRST		0x8
 #endif
 
 struct confparam {
 #ifdef INET6
 	int			inetdom;	/* internet domain */
 #endif
+	int			v4v6fallback;	/* fallbacking for v4/v6 only */
 	int			sd_max;		/* max of socket descriptor */
 	int			rcpt_min;	/* recipient lower limit */
 	int			sendbufsize;	/* TCP send buffer size */
 	int			data_max;	/* max size of message */
 	int			src_port;	/* SMTP source port */
+	int			cquery_max;	/* MAX concurrent queries */
 	int			dst_port;	/* SMTP destination port */
 	int			useAfirstMX;	/* use A RR if first MX */
 	int			nopipelining;	/* surpress SMTP pipelining */
@@ -389,14 +405,17 @@ struct confparam {
 	int			rcpts_trans;	/* recipients per transaction */
 	int			parsequeueid;	/* get queue-id info. */
 	int			showversion;	/* show version info. */
+	int			absentAok;	/* MXs point absent A are OK */
 	int			showrecipient;	/* show recipient info. */
 	int			smtp_reuse;	/* reuse SMTP connection */
 	time_t			t_timeout;	/* total timeout */
 	char			*map;		/* path of map file */
+	char			*pgateway;	/* name of protocol gateway */
 	char			*fallbackmx;	/* name of fallback MX host */
 	char			*sourceIP;	/* source address for SMTP */
 	char			*src_mod_path;	/* addr. list file to modify */
 	char			*src_mod_ptn;	/* modification pattern */
+	char			*rcpts_def;	/* path of file definded recipients of domain */
 };
 
 union SockAddrUnion {
