@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: dns.c,v 1.70 1999/10/16 07:55:15 motonori Exp $";
+static char *_id_ = "$Id: dns.c,v 1.84 2002/03/15 02:03:20 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -117,6 +117,30 @@ struct domain *dl;
 	else
 		setproctitle("pre-getting MX", NULL);
 
+	if (cnf.pgateway != NULL)
+	{
+		if (cnf.debug & DEBUG_DNS)
+		log(LOG_DEBUG, "pregetting A of protocol gateway: %s",
+			cnf.pgateway);
+#ifdef INET6
+		if (cnf.inetdom & SMTP_V4)
+		{
+			if (query_domain(cnf.pgateway, T_A) < 0)
+				return -1;
+			n++;
+		}
+		if (cnf.inetdom & SMTP_V6)
+		{
+			if (query_domain(cnf.pgateway, T_AAAA) < 0)
+				return -1;
+			n++;
+		}
+#else
+		if (query_domain(cnf.pgateway, T_A) < 0)
+			return -1;
+		n++;
+#endif
+	}
 	if (cnf.fallbackmx != NULL)
 	{
 		if (cnf.debug & DEBUG_DNS)
@@ -125,11 +149,9 @@ struct domain *dl;
 #ifdef INET6
 		if (cnf.inetdom & SMTP_V4)
 		{
-#endif
 			if (query_domain(cnf.fallbackmx, T_A) < 0)
 				return -1;
 			n++;
-#ifdef INET6
 		}
 		if (cnf.inetdom & SMTP_V6)
 		{
@@ -137,6 +159,10 @@ struct domain *dl;
 				return -1;
 			n++;
 		}
+#else
+		if (query_domain(cnf.fallbackmx, T_A) < 0)
+			return -1;
+		n++;
 #endif
 	}
 
@@ -164,11 +190,9 @@ struct domain *dl;
 #ifdef INET6
 				if (cnf.inetdom & SMTP_V4)
 				{
-#endif
 					if (query_domain(p, T_A) < 0)
 						return -1;
 					n++;
-#ifdef INET6
 				}
 				if (cnf.inetdom & SMTP_V6)
 				{
@@ -176,6 +200,10 @@ struct domain *dl;
 						return -1;
 					n++;
 				}
+#else
+				if (query_domain(p, T_A) < 0)
+					return -1;
+				n++;
 #endif
 			}
 			if (q != NULL)
@@ -190,6 +218,8 @@ struct domain *dl;
 
 	for (dp = dl; dp != NULL; dp = dp->next)
 	{
+		char *map_opt;
+
 		/* known by previous transaction */
 		if (dp->firstmx != NULL || dp->stat != EX_OK)
 			continue;
@@ -201,10 +231,14 @@ struct domain *dl;
 #endif
 
 		map_arg = dp->map_arg;
+		map_opt = strchr(*map_arg, '/');
 		while (*map_arg != NULL)
 		{
 			int len;
 
+			if (map_opt != NULL)
+				*map_opt = '\0'; /* ignore /opt portion */
+
 			if ((strcasecmp(*map_arg, "MX") == 0)
 			 || (strcasecmp(*map_arg, "MX?") == 0))
 			{
@@ -287,19 +321,29 @@ struct domain *dl;
   skip_preget_a:
 				(*map_arg)[len - 1] = ']';
 			}
+			if (map_opt != NULL)
+				*map_opt = '/';	/* restore /opt portion */
 			map_arg++;
 		}
 	}
-	if (env.queueid != NULL)
-		setproctitle("[%s] syncing MX: %d", env.queueid, n);
+	if (n > 0)
+	{
+		if (env.queueid != NULL)
+			setproctitle("[%s] syncing MX: %d", env.queueid, n);
+		else
+			setproctitle("syncing MX: %d", n);
+		if (cnf.debug & DEBUG_DNS)
+		log(LOG_DEBUG, "syncing responses of MXRRs");
+		if (process_query(1) < 0)	/* sync */
+			return -1;
+		etime = time(NULL);
+		log(LOG_DEBUG, "got MXRRs: %d MXs in %d seconds",
+			n, etime - stime);
+	}
 	else
-		setproctitle("syncing MX: %d", n);
-	if (cnf.debug & DEBUG_DNS)
-	log(LOG_DEBUG, "syncing responses of MXRRs");
-	if (process_query(1) < 0)	/* sync */
-		return -1;
-	etime = time(NULL);
-	log(LOG_DEBUG, "got MXRRs: %d MXs in %d seconds", n, etime - stime);
+	{
+		log(LOG_DEBUG, "no MXRRs to sync");
+	}
 	return 0;
 }
 
@@ -328,9 +372,7 @@ int type;
 	switch (type) {
 	  case T_MX:
 	  case T_A:
-#ifdef INET6
 	  case T_AAAA:
-#endif
 		break;
 	  default:
 		log(LOG_NOTICE, "query_domain: unknown query type: %d", type);
@@ -343,6 +385,7 @@ int type;
 
 static struct query_chain *QueryChain = NULL;
 static struct query_chain *RevQueryChain = NULL;
+static query_chain_count = 0;
 
 static int
 make_query(qrp, type)
@@ -362,11 +405,9 @@ int type;
 	  case T_A:
 		qflag = RR_A;
 		break;
-#ifdef INET6
 	  case T_AAAA:
 		qflag = RR_AAAA;
 		break;
-#endif
 	  default:
 		/* error */
 		return -1;
@@ -398,8 +439,7 @@ int type;
 	{
 		/* already in processing */
 		if (cnf.debug & DEBUG_DNS)
-		log(LOG_DEBUG, "skip request n=%s t=%d",
-			qrp->name, type);
+		log(LOG_DEBUG, "skip request n=%s t=%d", qrp->name, type);
 		return 0;
 	}
 
@@ -450,6 +490,11 @@ int type;
 		QueryChain = qchain;
 	}
 	qchain->prev = NULL;
+	query_chain_count++;
+	if (cnf.debug & DEBUG_DNS)
+	log(LOG_DEBUG, "DNS: query concurrency = %d", query_chain_count);
+	if (sti.maxcquery < query_chain_count)
+		sti.maxcquery = query_chain_count;
 
 #if 0
 	if (cnf.debug & DEBUG_DNS)
@@ -538,7 +583,7 @@ int sync;
         int max, n, fromlen, resplen, ns;
 	union {
 		HEADER  qb1;
-		char    qb2[PACKETSZ];
+		char    qb2[MAXPACKET];
 	} answer;
 	int anssiz;
         HEADER *qp, *hp;
@@ -549,6 +594,14 @@ int sync;
 	static void got_answer();
 	static void remove_query();
 
+	if (sync == 0 && cnf.cquery_max != 0
+	 && query_chain_count >= cnf.cquery_max)
+	{
+		if (cnf.debug & DEBUG_DNS)
+		log(LOG_DEBUG, "reached to max concurrency (%d)....syncing",
+			query_chain_count);
+		sync = -1;
+	}
 #if INET6
 	for (n = 0; n < _res.nscount; n++)
 	{
@@ -719,7 +772,7 @@ again:
 	do {
 		latest = 0;
 
-		if (sync) {
+		if (sync != 0) {
 			/* calculate next action time */
 			for (qcp = QueryChain; qcp != NULL; qcp = qcp->next)
 			{
@@ -896,7 +949,7 @@ again:
 		}
 	} while(n > 0);
 
-	if (sync) {
+	if (sync != 0) {
 		pending = 0;
 		if (cnf.debug & DEBUG_DNS)
 		log(LOG_DEBUG, "DNS: sync: get nearest timeout from %d", now);
@@ -920,6 +973,8 @@ again:
 			log(LOG_DEBUG, "DNS: event timeout=%d", qcp->time);
 			qcp = qcp->next;
 		}
+		if (sync < 0)
+			return (0); /* concurrency control */
 		if (pending)
 		{
 			if (cnf.debug & DEBUG_DNS)
@@ -978,6 +1033,7 @@ struct query_chain *qcp;
 		RevQueryChain = qcp->prev;
 	qcp->prev = NULL;
 	qcp->next = NULL;
+	query_chain_count--;
 
 #if 0
 	if (cnf.debug & DEBUG_DNS)
@@ -1028,6 +1084,14 @@ int anssiz;
 
 	ancount = ntohs(hp->ancount);
 
+	if (hp->aa == 0 && hp->ra == 0) {
+		/* this nameserver does not support recursion */
+		if (cnf.debug & DEBUG_DNS)
+		log(LOG_DEBUG, "non recursive ns=%d", qcp->ns);
+		qcp->badns |= 1<<qcp->ns;
+		return;
+	}
+
 	if (hp->rcode == NXDOMAIN
 	 || qcp->type == T_ANY && hp->rcode == NOERROR && ancount == 0)
 	{
@@ -1053,12 +1117,10 @@ int anssiz;
 			{
 				make_query(qcp->stat, T_MX);
 			}
-#ifdef INET6
 			else if (qcp->need == RR_AAAA)
 			{
 				make_query(qcp->stat, T_AAAA);
 			}
-#endif
 			else if (qcp->need == RR_A)
 			{
 				make_query(qcp->stat, T_A);
@@ -1070,15 +1132,19 @@ int anssiz;
 		{
 			/* no MX, try A/AAAA */
 #ifdef INET6
-			if (cnf.inetdom & SMTP_V6)
+			if (cnf.inetdom & SMTP_V6 || cnf.v4v6fallback)
 			{
 				make_query(qcp->stat, T_AAAA);
 			}
-			if (cnf.inetdom & SMTP_V4)
+			if (cnf.inetdom & SMTP_V4 || cnf.v4v6fallback)
 			{
-#endif
 				make_query(qcp->stat, T_A);
-#ifdef INET6
+			}
+#else
+			make_query(qcp->stat, T_A);
+			if (cnf.v4v6fallback)
+			{
+				make_query(qcp->stat, T_AAAA);
 			}
 #endif
 		}
@@ -1092,6 +1158,11 @@ int anssiz;
 	{
 		n = dn_expand((u_char *) b, eom, ap, nbuf, sizeof nbuf);
 		ap += n;
+		if (ap + INT16SZ*2 > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(type, ap);
 		GETSHORT(class, ap);
 	}
@@ -1107,15 +1178,33 @@ int anssiz;
 		if (n < 0)
 			return;
 		ap += n;
+		if (ap + INT16SZ*3 + INT32SZ > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(type, ap);
 		GETSHORT(class, ap);
-		ap += INT32SZ;
+		ap += INT32SZ; /* TTL */
 		GETSHORT(n, ap);
+		if (ap + n > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
+		if (class != C_IN)
+		{
+			if (cnf.debug & DEBUG_DNS)
+			log(LOG_DEBUG, "DNS: unexpected class %d in an", class);
+			ap += n;
+			continue;
+		}
 		if (type == T_MX)
 		{
 			if (strcasecmp(qcp->stat->name, nbuf) != 0)
 			{
 				/* may be a MX pointed by CNAME */
+				mx_found = 1;
 #if 1
 				cname_found = 0;
 #endif
@@ -1167,9 +1256,13 @@ int anssiz;
 						MAXDNAME+1);
 			}
 #endif
-#ifdef INET6
 		}
-		else if (type == T_AAAA && (cnf.inetdom & SMTP_V6))
+#ifdef INET6
+		else if (type == T_AAAA
+		      && ((cnf.inetdom & SMTP_V6) || cnf.v4v6fallback))
+#else
+		else if (type == T_AAAA && cnf.v4v6fallback)
+#endif
 		{
 			if (strcasecmp(qcp->stat->name, nbuf) == 0)
 			{
@@ -1197,16 +1290,21 @@ int anssiz;
 			}
 			bzero(&mxtmp, sizeof(mxtmp));
 			mxtmp.name = nbuf;
+#ifdef INET6
 			if (addinetaddress(&mxtmp, AF_INET6, n, ap) < 0)
 				return;
-		}
-		else if (type == T_A && (cnf.inetdom & SMTP_V4))
-		{
 #else
-		}
-		else if (type == T_A)
-		{
+			if (addinetaddress(&mxtmp, AF_UNSPEC, 0, "") < 0)
+				return;
 #endif
+		}
+#ifdef INET6
+		else if (type == T_A
+		      && ((cnf.inetdom & SMTP_V4) || cnf.v4v6fallback))
+#else
+		else if (type == T_A)
+#endif
+		{
 			if (strcasecmp(qcp->stat->name, nbuf) == 0)
 			{
 				addr_found = 1;
@@ -1323,11 +1421,9 @@ int anssiz;
 			  case RR_A:
 				make_query(qrp, T_A);
 				break;
-#ifdef INET6
 			  case RR_AAAA:
 				make_query(qrp, T_AAAA);
 				break;
-#endif
 			}
 		}
 	}
@@ -1343,7 +1439,17 @@ int anssiz;
 			goto noadditional;
 		ap += n;
 		ap += INT16SZ + INT16SZ + INT32SZ;
+		if (ap + INT16SZ > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(n, ap);
+		if (ap + n > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		ap += n;
 	}
 	while (--nscount >= 0 && ap < eom)
@@ -1352,7 +1458,17 @@ int anssiz;
 			goto noadditional;
 		ap += n;
  		ap += INT16SZ + INT16SZ + INT32SZ;
+		if (ap + INT16SZ > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(n, ap);
+		if (ap + n > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		ap += n;
 	}
 	while (--arcount >= 0 && ap < eom)
@@ -1363,9 +1479,27 @@ int anssiz;
 				nbuf, sizeof nbuf)) < 0)
 			return;
 		ap += n;
+		if (ap + INT16SZ*3 + INT32SZ > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(type, ap);
- 		ap += INT16SZ + INT32SZ;
+		GETSHORT(class, ap);
+ 		ap += INT32SZ; /* TTL */
 		GETSHORT(addrsize, ap);
+		if (ap + addrsize > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
+		if (class != C_IN)
+		{
+			if (cnf.debug & DEBUG_DNS)
+			log(LOG_DEBUG, "DNS: unexpected class %d in ar", class);
+			ap += addrsize;
+			continue;
+		}
 		switch (type)
 		{
 #ifdef INET6
@@ -1445,8 +1579,10 @@ int anssiz;
 
 	if (mx_found)
 	{
-
-		mxp = qcp->stat->mx;
+		qrp = qcp->stat;
+		while (qrp->cname != NULL)
+			qrp = qrp->cname;
+		mxp = qrp->mx;
 		while (mxp != NULL)
 		{
 			qrp = hash_query_lookup(mxp->name, &hashp);
@@ -1466,21 +1602,28 @@ int anssiz;
 			}
 
 #ifdef INET6
-			if (cnf.inetdom & SMTP_V6)
+			if (cnf.inetdom & SMTP_V6 || cnf.v4v6fallback)
 			{
 				if (cnf.debug & DEBUG_DNS)
 				log(LOG_DEBUG, "DNS: getting AAAA RR for %s",
 					mxp->name);
 				make_query(qrp, T_AAAA);
 			}
-			if (cnf.inetdom & SMTP_V4)
+			if (cnf.inetdom & SMTP_V4 || cnf.v4v6fallback)
 			{
-#endif
 				if (cnf.debug & DEBUG_DNS)
 				log(LOG_DEBUG, "DNS: getting A RR for %s",
 					mxp->name);
 				make_query(qrp, T_A);
-#ifdef INET6
+			}
+#else
+			if (cnf.debug & DEBUG_DNS)
+			log(LOG_DEBUG, "DNS: getting A RR for %s",
+				mxp->name);
+			make_query(qrp, T_A);
+			if (cnf.v4v6fallback)
+			{
+				make_query(qrp, T_AAAA);
 			}
 #endif
 			mxp = mxp->next;
@@ -1498,8 +1641,9 @@ getmxlist(dl)
 struct domain *dl;
 {
 	struct domain *dp;
-	char **map_arg;
+	char **map_arg, *map_opt, *p;
 	int offset, needfindaddr, rc;
+	long opt;
 	struct mx *mxp;
 
 	for (dp = dl; dp != NULL; dp = dp->next)
@@ -1516,19 +1660,40 @@ struct domain *dl;
 		map_arg = dp->map_arg;
 		while (*map_arg != NULL)
 		{
+			map_opt = strchr(*map_arg, '/');
+			if (map_opt != NULL)
+				*map_opt++ = '\0'; /* split /opt portion */
+			opt = 0;
+			if ((p = map_opt) != NULL)
+			{
+				while (*p != '\0')
+				{
+					switch (*p)
+					{
+					  case 'a':
+						opt |= MAPOPT_TRYALLADDR;
+						break;
+					}
+					p++;
+				}
+			}
+
 			if (strcasecmp(*map_arg, "MX") == 0)
-				rc = getmxrr(dp->name, dp, 1, offset, 1);
+				rc = getmxrr(dp->name, dp, 1, offset, 1, opt);
 			else if (strcasecmp(*map_arg, "MX?") == 0)
-				rc = getmxrr(dp->name, dp, 1, offset, 0);
+				rc = getmxrr(dp->name, dp, 1, offset, 0, opt);
 			else if (strcasecmp(*map_arg, "A") == 0)
-				rc = getmxrr(dp->name, dp, 0, offset, 0);
+				rc = getmxrr(dp->name, dp, 0, offset, 0, opt);
 			else
-				rc = getmxrr(*map_arg, dp, 1, offset, 0);
+				rc = getmxrr(*map_arg, dp, 1, offset, 0, opt);
 			if (rc < 0 || dp->stat != EX_OK)
 				goto next_domain;
 			if (rc > 0)
 				needfindaddr = 1;
 			offset += 300;
+
+			if (map_opt != NULL)
+				*--map_opt = '/'; /* restore */
 			map_arg++;
 		}
 
@@ -1557,10 +1722,70 @@ struct domain *dl;
 				if (mxp->host == NULL)
 				{
 					/* set address information */
-					addinetaddress(mxp, 0, 0, NULL);
+					addinetaddress(mxp, AF_UNSPEC, 0, NULL);
 				}
 			}
 		}
+
+		if (cnf.pgateway != NULL && cnf.v4v6fallback)
+		{
+			struct inetaddr *addrp;
+			int found = 0;
+
+			if (cnf.debug & DEBUG_DNS)
+			log(LOG_DEBUG, "protocol gateway checking for %s",
+				dp->name);
+			for (mxp = dp->firstmx; mxp != NULL; mxp = mxp->next)
+			{
+				if (cnf.debug & DEBUG_DNS)
+				log(LOG_DEBUG, "checking MX: %s", mxp->name);
+				if (mxp->host == NULL)
+				{
+					/* no hosts, try next MX */
+					mxp = mxp->next;
+					continue;
+				}
+				for(addrp = mxp->host->addr; addrp != NULL;
+				    addrp = addrp->next)
+				{
+/*
+					if (cnf.debug & DEBUG_DNS)
+					log(LOG_DEBUG, "proto: %d",
+						addrp->domain);
+*/
+#ifdef INET6
+					if(((cnf.inetdom & SMTP_V4) == 0
+					 && addrp->domain == AF_INET)
+					|| ((cnf.inetdom & SMTP_V6) == 0
+					 && addrp->domain == AF_INET6))
+#else
+					if(addrp->domain != AF_INET)
+#endif
+					{
+						struct mx *gwmxp;
+						static struct mx* newmx();
+
+						if (cnf.debug & DEBUG_DNS)
+						log(LOG_DEBUG, "setting protocol gateway as first MX");
+						if ((gwmxp = newmx(cnf.pgateway,
+								  -2, 0, dp, 0))
+						    == NULL)
+						{
+							/* out of memory */
+							found = 1;
+							break;
+						}
+						/* set address information */
+						addinetaddress(gwmxp, AF_UNSPEC,
+							       0, NULL);
+						found = 1;
+						break;
+					}
+				}
+				if (found)
+					break;
+			}
+		}
   next_domain:;
 	}
 	return 0;
@@ -1609,10 +1834,11 @@ struct mx *newmxp;
 }
 
 static struct mx*
-newmx(name, pref, weight, dp)
+newmx(name, pref, weight, dp, opt)
 char *name;
 int pref, weight;
 struct domain *dp;
+long opt;
 {
 	struct mx *mxp;
 	struct mx *pmxp = NULL;	/* just for warning suppression */
@@ -1672,6 +1898,7 @@ struct domain *dp;
 		mxp->pref = pref;
 		mxp->weight = weight;
 		mxp->domain = dp;
+		mxp->opt = opt;
 	}
 
 	linkmx(dp, mxp);
@@ -1680,12 +1907,13 @@ struct domain *dp;
 }
 
 int
-getmxrr(name, dp, query_mx, offset, wait)
+getmxrr(name, dp, query_mx, offset, wait, opt)
 char *name;
 struct domain *dp;
 int query_mx;
 int offset;
 int wait;	/* wait till next trial if MX lookup failure */
+long opt;	/* option described in map file */
 {
 	int seenlocal, localpref, seenfallback;
 	struct mx *mxp, *pmxp, *cmxp;
@@ -1694,12 +1922,13 @@ int wait;	/* wait till next trial if MX lookup failure */
 	char msgbuf[MAXLINE], *mxname;
 
 	if (cnf.debug & DEBUG_DNS)
-	log(LOG_DEBUG, "getmxrr(%s, %s) offset=%d", name, dp->name, offset);
+	log(LOG_DEBUG, "getmxrr(%s, %s) offset=%d opt=%x",
+		name, dp->name, offset, opt);
 
 	if (!query_mx)
 	{
 		/* A RR lookup */
-		mxp = newmx(name, offset - 1, -1, dp);
+		mxp = newmx(name, offset - 1, -1, dp, opt);
 		if (mxp == NULL)
 			return -1;
 		return 1;	/* find adderess */
@@ -1728,7 +1957,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 		{
 			/* numeric IPv4 address */
 			name[len-1] = ']';
-			mxp = newmx(name, offset - 1, -1, dp);
+			mxp = newmx(name, offset - 1, -1, dp, opt);
 			if (mxp == NULL)
 				return -1;
 			addinetaddress(mxp, AF_INET, sizeof(nsin.in.sin_addr),
@@ -1739,7 +1968,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 		{
 			/* numeric IPv6 address */
 			name[len-1] = ']';
-			mxp = newmx(name, offset - 1, -1, dp);
+			mxp = newmx(name, offset - 1, -1, dp, opt);
 			if (mxp == NULL)
 				return -1;
 			addinetaddress(mxp, AF_INET6,
@@ -1750,7 +1979,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 		else
 		{
 			/* A RR lookup */
-			mxp = newmx(name + 1, offset - 1, -1, dp);
+			mxp = newmx(name + 1, offset - 1, -1, dp, opt);
 			name[len-1] = ']';
 			if (mxp == NULL)
 				return -1;
@@ -1788,7 +2017,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 		if (cnf.fallbackmx != NULL)
 		{
 			dp->stat = EX_OK;
-			mxp = newmx(cnf.fallbackmx, offset - 1, 0, dp);
+			mxp = newmx(cnf.fallbackmx, offset - 1, 0, dp, opt);
 			if (mxp == NULL)
 				return -1;
 			return 1;	/* adderess found */
@@ -1824,9 +2053,15 @@ int wait;	/* wait till next trial if MX lookup failure */
 		/* no MX on this host, may be A exists */
 		dp->stat = EX_OK;
 		/* give error if A for the host is not found */
-		mxp = newmx(qrp->name, offset - 1, -1, dp);
+		mxp = newmx(qrp->name, offset - 1, -1, dp, opt);
 		if (mxp == NULL)
 			return -1;
+		if (cnf.fallbackmx != NULL)
+		{
+			mxp = newmx(cnf.fallbackmx, offset - 1, 0, dp, opt);
+			if (mxp == NULL)
+				return -1;
+		}
 		return 1;	/* find adderess */
 	}
 
@@ -1835,9 +2070,15 @@ int wait;	/* wait till next trial if MX lookup failure */
 		/* may be CNAME tail, no MX on this host, may be A exists */
 		dp->stat = EX_OK;
 		/* give error if A for the host is not found */
-		mxp = newmx(qrp->name, offset - 1, -1, dp);
+		mxp = newmx(qrp->name, offset - 1, -1, dp, opt);
 		if (mxp == NULL)
 			return -1;
+		if (cnf.fallbackmx != NULL)
+		{
+			mxp = newmx(cnf.fallbackmx, offset - 1, 0, dp, opt);
+			if (mxp == NULL)
+				return -1;
+		}
 		return 1;	/* find adderess */
 	}
 
@@ -1916,7 +2157,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 		}
 
 		mxp = newmx(mxname, offset + cmxp->pref,
-			mxrand(mxname, qrp->name), dp);
+			mxrand(mxname, qrp->name), dp, opt);
 		if (mxp == NULL)
 			return -1;
 	}
@@ -1950,7 +2191,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 	{
 		if (cnf.useAfirstMX)
 		{
-			mxp = newmx(qrp->name, offset - 1, -1, dp);
+			mxp = newmx(qrp->name, offset - 1, -1, dp, opt);
 			if (mxp == NULL)
 				return -1;
 		}
@@ -1964,7 +2205,7 @@ int wait;	/* wait till next trial if MX lookup failure */
 
 	if (cnf.fallbackmx != NULL && !seenfallback)
 	{
-		mxp = newmx(cnf.fallbackmx, offset + 256, 0, dp);
+		mxp = newmx(cnf.fallbackmx, offset + 256, 0, dp, opt);
 		if (mxp == NULL)
 			return -1;
 	}
@@ -2003,6 +2244,11 @@ u_char *t, *b, *e;
 	{
 		n = dn_expand((u_char *) b, eom, ap, nbuf, sizeof nbuf);
 		ap += n;
+		if (ap + INT16SZ*2 > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(class, ap);
 		GETSHORT(type, ap);
 		if (cnf.debug & DEBUG_DNS)
@@ -2019,10 +2265,27 @@ u_char *t, *b, *e;
 		if (n < 0)
 			break;
 		ap += n;
+		if (ap + INT16SZ*3 + INT32SZ > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
 		GETSHORT(class, ap);
 		GETSHORT(type, ap);
-		ap += INT32SZ;
+		ap += INT32SZ; /* TTL */
 		GETSHORT(n, ap);
+		if (ap + n > eom)
+		{
+			/* log(LOG_NOTICE, "corrupt DNS answer"); */
+			return;
+		}
+		if (class != C_IN)
+		{
+			if (cnf.debug & DEBUG_DNS)
+			log(LOG_DEBUG, "DNS: unexpected class %d in an", class);
+			ap += n;
+			continue;
+		}
 		if (type == T_MX) {
 			GETSHORT(pref, ap);
 			if ((n = dn_expand((u_char *) b, eom, ap,
