@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,7 +33,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: extern.h,v 1.18 1999/03/02 19:28:48 motonori Exp $
+ * $Id: extern.h,v 1.24 2002/01/20 14:20:42 motonori Exp $
  */
 
 char *newstr();
@@ -62,6 +62,11 @@ struct host_map *hash_map_lookup();
 int host_map_load();
 char **host_map_lookup();
 int mxrand();
+int hash_domain_rcpts_lookup();
+void hash_domain_rcpts_enter();
+int isnumeric();
+int rcpts_def_load();
+
 void initsetproctitle();
 #ifdef JAZ_HACK2
 void show_cache();
