@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: lmtp.c,v 1.56 1999/02/19 00:35:40 motonori Exp $";
+static char *_id_ = "$Id: lmtp.c,v 1.61 2002/01/20 14:19:39 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -98,7 +98,7 @@ lmtp()
 
 	save_rcpts_trans = cnf.rcpts_trans;
 
-	setproctitle("LMTP: waiting message", NULL);
+	setproctitle("LMTP: waiting greeting", NULL);
 	lmtp_greeting();
 	fflush (LmtpOut);
 	state = LMTP_INIT;
@@ -344,6 +344,7 @@ char *param;
 	fprintf (LmtpOut,
 		"354 Enter mail, end with \".\" on a line by itself\n");
 	fflush (LmtpOut);
+	setproctitle("LMTP: reading a message", NULL);
 	if (read_message() < 0) {
 		/* to fallback next mailer (sendmail internal SMTP) */
 		fprintf (LmtpOut, "421 message too large (no more memory)\n");
@@ -357,10 +358,11 @@ char *param;
 	deliver();		/* delivery with SMTP */
 	lmtp_data_response(1);
 
-	log(LOG_INFO, "nquery=%d/%d nconnect=%d/%d ntimeout=%d ntrans=%d/%d nsent=%d/%d ndeferred=%d nerror=%d+%d maxsock=%d tmx=%d tdelivery=%d/%d maxdelay=%d tdelay=%d tsendmail=%d dns_i/o=%d/%d smtp_i/o=%d/%d mem=%d",
+	log(LOG_INFO, "nquery=%d/%d nconnect=%d/%d ntimeout=%d ntrans=%d/%d nsent=%d/%d ndeferred=%d nerror=%d+%d dnsmaxcq=%d maxsock=%d tmx=%d tdelivery=%d/%d maxdelay=%d tdelay=%d tsendmail=%d dns_i/o=%d/%d smtp_i/o=%d/%d mem=%d",
 		sti.nanswers, sti.nqueries, sti.nquitok, sti.nconnect,
 		sti.ntimeout, sti.noktrans, sti.ntrans, sti.nsent, sti.nrcpt,
-		sti.ndeferred, sti.nnsfailed, sti.nsmtpfailed, sti.maxsock,
+		sti.ndeferred, sti.nnsfailed, sti.nsmtpfailed,
+		sti.maxcquery, sti.maxsock,
 		sti.time_mxgot - sti.time_start,
 		sti.time_almostfinish - sti.time_mxgot,
 		sti.time_finish - sti.time_mxgot,
