@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -39,7 +39,7 @@ static char *copyright1 = "@(#)Copyright (c) 1997,1998 Motonori Nakamura";
 static char *copyright2 = "@(#)Copyright (c) 1997,1998 WIDE Project.";
 static char *copyright3 = "@(#)Copyright (c) 1997,1998 Kyoto University.";
 static char *copyright4 = "@(#)All rights reserved.";
-static char *_id_ = "$Id: main.c,v 1.55 1999/04/15 05:57:10 motonori Exp $";
+static char *_id_ = "$Id: main.c,v 1.73 2002/03/15 02:12:52 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -53,6 +53,7 @@ char *progname;
 	printf("Usage: %s [options]\n", progname);
 	printf("Options are:\n");
 	printf("	-1		no piggybacking; par recipient SMTP\n");
+	printf("	-5		check A/AAAA to fallback for v4 or v6 only\n");
 	printf("	-a alias	my alias names to check MX list\n");
 	printf("	-A		try T_ANY query first to get MX\n");
 	printf("	-b size		TCP send buffer size\n");
@@ -65,6 +66,7 @@ char *progname;
 	printf("\t\t	D: no delivery	Q: DNS query only\n");
 	printf("	-E		get queue-id info. from Received:\n");
 	printf("	-F		insert recipient info. on Received:\n");
+	printf("	-g gateway	set protocol gateway MX host\n");
 	printf("	-h		show this message\n");
 	printf("	-I queue-id	equiv. to -i but no purge on RSET\n");
 	printf("	-i queue-id	queue-id from sendmail for logging\n");
@@ -73,11 +75,13 @@ char *progname;
 	printf("	-m fallbackmx	set fallback MX host\n");
 #ifdef INET6
 	printf("	-N domain	4:IPv4 only, 6:IPv6 only, B:Both\n");
+	printf("\t		B4:try IPv4 first, B6:try IPv6 first\n");
 #endif
 	printf("	-n num		available number of sockets\n");
 	printf("	-o source-addr	source address for SMTP connection\n");
 	printf("	-p port		SMTP source port number\n");
-	printf("	-R num		max recipients per transaction\n");
+	printf("	-q num		max number of concurrent queries\n");
+	printf("	-R num/file	max recipients per transaction\n");
 	printf("	-r num		min recipients to work\n");
 	printf("	-s size		max message size\n");
 	printf("	-t cate=val	set timeouts (val: NUM[hms]; default is s)\n");
@@ -85,6 +89,7 @@ char *progname;
 	printf("\t		rcpt, data, body, term, cache, rset, quit\n");
 	printf("	-u		reuse SMTP connection for other MX\n");
 	printf("	-V		insert version info. on Received:\n");
+	printf("	-x		no error even if MX points absent A\n");
 	printf("	-y\t	refer A RR if this host is the first MX\n");
 	printf("	-z		surpress SMTP pipelining\n");
 	printf("Version: %s\n", version);
@@ -96,7 +101,7 @@ sig_quit()
 	exit(EX_TEMPFAIL);
 }
 
-void
+int
 main(argc, argv, envp)
 int argc;
 char *argv[];
@@ -109,6 +114,12 @@ char *envp[];
 	char *progname, *p;
 	int opt;
 	struct rlimit limit;
+	SockAddr peername;
+	int peernamelen;
+#define SHOW_UID 0
+#if SHOW_UID
+	uid_t uid;
+#endif
 
 	if (chdir("/var/tmp") < 0)
 		if (chdir("/usr/tmp") < 0)
@@ -136,6 +147,7 @@ char *envp[];
 	cnf.sd_max = MAX_SOCK;
 	cnf.log_facility = LOG_DEFAULT;
 	cnf.rcpts_trans = 100;
+	cnf.cquery_max = 50;
 	cnf.src_mod_ptn = "<%s+%s=%s@%s>";
 
         if((sp = getservbyname("smtp", "tcp")) != NULL)
@@ -147,9 +159,9 @@ char *envp[];
 
 	opterr = 0;
 #ifdef INET6
-#define	OPTS	"1Aa:b:c:d:EFhI:i:l:M:m:N:n:o:p:R:r:S:s:t:uVyz"
+#define	OPTS	"15Aa:b:c:d:EFg:hI:i:l:M:m:N:n:o:p:q:R:r:S:s:t:uVxyz"
 #else
-#define	OPTS	"1Aa:b:c:d:EFhI:i:l:M:m:n:o:p:R:r:S:s:t:uVyz"
+#define	OPTS	"15Aa:b:c:d:EFg:hI:i:l:M:m:n:o:p:R:q:r:S:s:t:uVxyz"
 #endif
 	while ((opt = getopt(argc, argv, OPTS)) != -1)
 	{
@@ -158,6 +170,9 @@ char *envp[];
 		    case '1':	/* 1: no piggybacking */
 			cnf.rcpts_trans = 1;
 			break;
+		    case '5':	/* 5: check A/AAAA to fallback for v4/v6 only */
+			cnf.v4v6fallback = 1;
+			break;
 		    case 'A':	/* A: try T_ANY first */
 			cnf.tryANYfirst = 1;
 			break;
@@ -232,6 +247,11 @@ char *envp[];
 		    case 'F':	/* F: show recipient information */
 			cnf.showrecipient++;
 			break;
+		    case 'g':	/* g: protocol gateway MX */
+			if (cnf.pgateway != NULL)
+				free(cnf.pgateway);
+			cnf.pgateway = newstr(optarg);
+			break;
 		    case 'I':	/* I: queue-id just for logging (no purge) */
 			if (env.queueid != NULL)
 				free(env.queueid);
@@ -316,6 +336,15 @@ char *envp[];
 			  case 'b':
 			  case 'B':
 				cnf.inetdom |= SMTP_V4|SMTP_V6;
+				switch (optarg[1])
+				{
+				  case '4':
+					cnf.inetdom |= SMTP_V4_FIRST;
+					break;
+				  case '6':
+					cnf.inetdom |= SMTP_V6_FIRST;
+					break;
+				}
 				break;
 			}
 			break;
@@ -334,9 +363,22 @@ char *envp[];
 				/* rresvport() used if negative */
 			cnf.src_port = atoi(optarg);
 			break;
-		    /* q: quick start */
+		    /* Q: quick start */
+		    case 'q':	/* q: max number of concurrent queries */
+			cnf.cquery_max = atoi(optarg);
+			break;
 		    case 'R':	/* R: recipients par transaction */
+#if 0
 			cnf.rcpts_trans = atoi(optarg);
+#else
+			if (cnf.rcpts_def != NULL)
+				free(cnf.rcpts_def);
+			cnf.rcpts_def = NULL;
+			if(isnumeric(optarg))
+				cnf.rcpts_trans = atoi(optarg);
+			else
+				cnf.rcpts_def = newstr(optarg);
+#endif
 			break;
 		    case 'r':	/* r: number of recipients lower limit */
 			cnf.rcpt_min = atoi(optarg);
@@ -369,6 +411,9 @@ char *envp[];
 		    case 'V':	/* V: show version info. on Received: */
 			cnf.showversion = 1;
 			break;
+		    case 'x':	/* x: no error even if MX points absent A */
+			cnf.absentAok = 1;
+			break;
 		    case 'y':	/* y: refer A RR if this host is first MX */
 			cnf.useAfirstMX = 1;
 			break;
@@ -396,11 +441,33 @@ char *envp[];
 #ifdef INET6
 	/* set default */
 	if (cnf.inetdom == 0)
-		cnf.inetdom |= SMTP_V4|SMTP_V6;
+		cnf.inetdom |= SMTP_V4|SMTP_V6|SMTP_V6_FIRST;
 #endif
 
 	init_log();
 
+	peernamelen = sizeof(peername);
+	if (getpeername(0, (struct sockaddr *)&peername, &peernamelen) == 0)
+	{
+#ifdef INET6
+		char buf[MAXLINE];
+#endif
+		switch (peername.in.sin_family)
+		{
+		    case AF_INET:
+			log(LOG_INFO, "connected from %s",
+				inet_ntoa(peername.in.sin_addr));
+			break;
+#ifdef INET6
+		    case AF_INET6:
+			inet_ntop(AF_INET6, &peername.in6.sin6_addr,
+				buf, sizeof(buf));
+			log(LOG_INFO, "connected from %s", buf);
+			break;
+#endif
+		}
+	}
+
 	if (cnf.sd_max > MAX_SOCK)
 	{
 		log(LOG_INFO, "cnf.sd_max should not be greater than %d, recompile smtpfeed with greater FD_SETSIZE", MAX_SOCK);
@@ -424,11 +491,23 @@ char *envp[];
 	}
 #endif
 
+#if SHOW_UID
+	uid = geteuid();
+#ifdef INET6
+        log(LOG_DEBUG, "%s (with IPv6) started (uid %d with %d descriptors)",
+		version, uid, cnf.sd_max);
+#else
+        log(LOG_DEBUG, "%s started (uid %d with %d descriptors)",
+		version, uid, cnf.sd_max);
+#endif
+#else
 #ifdef INET6
         log(LOG_DEBUG, "%s (with IPv6) started (with %d descriptors)",
 		version, cnf.sd_max);
 #else
-        log(LOG_DEBUG, "%s started (with %d descriptors)", version, cnf.sd_max);
+        log(LOG_DEBUG, "%s started (with %d descriptors)",
+		version, cnf.sd_max);
+#endif
 #endif
 
 	res_init();
@@ -484,6 +563,22 @@ char *envp[];
 		free(cnf.fallbackmx);
 		cnf.fallbackmx = NULL;
 	}
+	if (cnf.pgateway != NULL && isamyalias(cnf.pgateway))
+	{
+		log(LOG_INFO, "ignoring protocol gateway (%s)", cnf.pgateway);
+		free(cnf.pgateway);
+		cnf.pgateway = NULL;
+	}
+
+	if (cnf.rcpts_def != NULL)
+	{
+		if(rcpts_def_load() < 0)
+		{
+			log(LOG_INFO, "ignoring R option (%s)", cnf.rcpts_def);
+			free(cnf.rcpts_def);
+			cnf.rcpts_def = NULL;
+		}
+	}
 
 	/* stdin/stdout will be used */
 #if 0
@@ -499,4 +594,6 @@ char *envp[];
 	resource_usage("all done");
 
 	exit(EX_OK);
+	/*NOTREACHED*/
+	return EX_OK;
 }
