@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: smtp.c,v 1.130 1999/04/15 05:43:17 motonori Exp $";
+static char *_id_ = "$Id: smtp.c,v 1.149 2002/02/23 10:19:49 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -226,10 +226,10 @@ static struct	{
 	{smtp_helo_r,	smtp_flush,	 5 MINUTES, 	"EHLOSENT"},	/* 3 */
 	{smtp_helo_r,	smtp_flush,	 5 MINUTES,	"HELOSENT"},	/* 4 */
 	{smtp_mail_r,	smtp_flush,	10 MINUTES,	"MAILSENT"},	/* 5 */
-	{smtp_rcpt_r,	smtp_rcpt_s,	60 MINUTES,	"RCPTSENT"},	/* 6 */
+	{smtp_rcpt_r,	smtp_rcpt_s,	10 MINUTES,	"RCPTSENT"},	/* 6 */
 	{smtp_data_r,	smtp_flush,	 5 MINUTES,	"DATASENT"},	/* 7 */
-	{NULL,		smtp_data_body,	20 MINUTES,	"DATABODY"},	/* 8 */
-	{smtp_data_t,	smtp_flush,	60 MINUTES,	"DTERMSENT"},	/* 9 */
+	{NULL,		smtp_data_body,	15 MINUTES,	"DATABODY"},	/* 8 */
+	{smtp_data_t,	smtp_flush,	20 MINUTES,	"DTERMSENT"},	/* 9 */
 	{NULL,		NULL,		 5 MINUTES,	"SENT"},	/* 10 */
 	{smtp_rset_r,	smtp_flush,	 5 MINUTES,	"RSETSENT"},	/* 11 */
 	{smtp_quit_r,	smtp_flush,	 2 MINUTES,	"QUITSENT"},	/* 12 */
@@ -585,6 +585,7 @@ new_transaction()
 	struct domain *domp;
 	struct mx *lastgoodmx;
 	struct connection *conn;
+	int no_addr;
 	int rc;
 
 	if (env.work_dom_ptr == NULL)
@@ -595,18 +596,50 @@ new_transaction()
 		if (cnf.debug & DEBUG_TRANS)
 		log(LOG_DEBUG, "checking domain: %s", domp->name);
 #endif
+		/* first time for the domain? */
+#if 1
+		if (domp->response == NULL)
+#else
+		if (domp->firstmx == domp->curmx)
+#endif
+			no_addr = 1;
+		else
+			no_addr = 0;
+
 		/* skip bad MXs */
 		lastgoodmx = NULL;
 		while (domp->curmx != NULL
 		    && (domp->curmx->host == NULL
 		     || (domp->curmx->host != NULL
-		     && domp->curmx->host->state == STAT_FAIL)))
+		     && (domp->curmx->host->state == STAT_FAIL
+		      || domp->curmx->host->addr == NULL))))
 		{
 			if (domp->curmx->host == NULL)
 				log(LOG_INFO, "skipping bad MX: %s for %s",
 					domp->curmx->name, domp->name);
 			else
+			{
+				struct dns_stat *qrp;
+
 				lastgoodmx = domp->curmx;
+				if (domp->curmx->host->addr != NULL)
+					no_addr = 0;
+				else
+				{
+					if (cnf.absentAok)
+						qrp = NULL;
+					else
+						qrp = hash_query_lookup(domp->curmx->host->name, NULL);
+#ifdef INET6
+					if (qrp == NULL
+					 || (qrp->rr_noex & (RR_A | RR_AAAA)) == 0)
+#else
+					if (qrp == NULL
+					 || (qrp->rr_noex & RR_A) == 0)
+#endif
+						no_addr = 0;
+				}
+			}
 			domp->curmx = domp->curmx->next;
 		}
 		if (domp->curmx == NULL)
@@ -614,7 +647,10 @@ new_transaction()
 			char *msg = domp->response;
 
 			if (msg == NULL)
-				msg = "No MX available";
+				if (no_addr)
+					msg = "No addresses for MX";
+				else
+					msg = "No MX available";
 			rcptp = domp->rcpt_top;
 			domp->curmx = lastgoodmx; /* just for markstatus */
 			while (rcptp != NULL)
@@ -622,8 +658,11 @@ new_transaction()
 				if (rcptp->stat == RCPT_NOOP
 				 || rcptp->stat == RCPT_RETRY)
 				{
-					markstatus(rcptp, SMTP_TEMPFAIL(52),
-						msg, 0);
+					if (no_addr)
+						rc = SMTP_ERROR(52);
+					else
+						rc = SMTP_TEMPFAIL(52);
+					markstatus(rcptp, rc, msg, 0);
 					rcptp->stat = RCPT_RETRY;
 				}
 				rcptp = rcptp->dom_chain;
@@ -673,15 +712,35 @@ new_transaction()
 
 	if (conn == NULL)
 		return -1;	/* XXX */
-#if 1
+#if 0
 	if (conn->host->addr == NULL)
 	{
 		char msgbuf[MAXLINE];
+		struct dns_stat *qrp;
+		int rcode;
 
 		snprintf(msgbuf, sizeof(msgbuf),
 			"No address known for %s", conn->host->name);
 		log(LOG_INFO, msgbuf, conn->host->name);
-		markstatus(conn->rcpt, SMTP_TEMPFAIL(52), msgbuf, 1);
+
+		rcode = SMTP_TEMPFAIL(52);
+		if (cnf.absentAok)
+			qrp = NULL;
+		else
+			qrp = hash_query_lookup(conn->host->name, NULL);
+		if (qrp != NULL)
+		{
+			while (qrp->cname != NULL)
+				qrp = qrp->cname;
+#ifdef INET6
+			if (qrp->rr_noex & (RR_A | RR_AAAA))
+#else
+			if (qrp->rr_noex & RR_A)
+#endif
+				rcode = SMTP_ERROR(50);
+		}
+		markstatus(conn->rcpt, rcode, msgbuf, 1);
+
 		close_transaction(conn, 1);
 		return 0;
 	}
@@ -841,9 +900,17 @@ int tracechain;
 				sti.max_delay = delay;
 			pbuf[0] = '\0';
 #ifdef INET6
-			if (hostp->current != NULL
-			 && hostp->current->domain == AF_INET6)
-				strncat(pbuf, "v6", sizeof(pbuf));
+			if (hostp->current != NULL)
+				switch (hostp->current->domain)
+				{
+				  case AF_INET6:
+					strncat(pbuf, "v6", sizeof(pbuf));
+					break;
+
+				  case AF_INET:
+					strncat(pbuf, "v4", sizeof(pbuf));
+					break;
+				}
 			if (hostp->proto & PROTO_ESMTP)
 				strncat(pbuf, (*pbuf != '\0')?",esmtp":"esmtp",
 					sizeof(pbuf));
@@ -894,10 +961,10 @@ int tracechain;
 				inet_ntop(AF_INET6,
 					hostp->current->address,
 					buf, sizeof(buf));
-				snprintf(hbuf, sizeof(hbuf), "%s [IPv6:%s]",
+				snprintf(hbuf, sizeof(hbuf), "%s [%s]",
 					hostp->name, buf);
 				snprintf(rbuf, sizeof(rbuf),
-					"%s (%s [IPv6:%s])",
+					"%s (%s [%s])",
 					response, hostp->name, buf);
 				break;
 #endif
@@ -917,6 +984,10 @@ int tracechain;
 					"%s (%s [unknown domain])",
 					response, hostp->name);
 			}
+			if (hostp->name != NULL
+			 && cnf.pgateway != NULL
+			 && strcasecmp(hostp->name, cnf.pgateway) == 0)
+				strncat(hbuf, " (gateway)", sizeof(hbuf));
 		}
 		else
 		{
@@ -952,12 +1023,33 @@ int tracechain;
 		{
 			if (tracechain == 0)
 				break;
+			/* skip alreadly marked recipient */
 			rcptp = rcptp->chain;
 			continue;
 		}
 		rcptp->result = rcode;
 		rcptp->response = responsep;
 
+#define SYSLOG_BUFFREE	( SYSLOG_BUFSIZE \
+			- 48 /* time stamp, hostname, pid, etc. */ \
+			- 64 /* constant string size (relay= etc.) */ \
+			- estconst \
+			- addrlen \
+			- 10 ) /*,(more...)*/
+
+		if (tracechain && strlen(rcptp->address) >= SYSLOG_BUFFREE)
+		{
+			/* buffer may be full */
+			strcat(addrbuf+addrlen, ",(more...)");
+			log(LOG_INFO, "(%d+%d+%d+%d/%d) relay=%s to=%s proto=%s delay=%d code=%d (%s)",
+				sti.nsent, sti.ndeferred, sti.nnsfailed,
+				sti.nsmtpfailed, sti.nrcpt, hostname,
+				addrbuf, proto, delay, rcode,
+				(response==NULL)?"?":response);
+			addrlen = 0;
+			addrbuf[0] = '\0';
+		}
+
 		if (IS2xx(rcode))
 		{
 			rcptp->stat = RCPT_DONE;
@@ -1005,35 +1097,18 @@ int tracechain;
 			addrlen += strlen(rcptp->address);
 		}
 
-		/* length estimation with next recipient */
-		if (tracechain && rcptp->chain != NULL)
-		{
-			estlen = 48 /* time stamp, hostname, pid, etc. */
-			       + 64 /* constant string size (relay= etc.) */
-			       + estconst + addrlen
-			       + strlen(rcptp->chain->address)
-			       + ((rcptp->chain->chain != NULL)?10:0);
-		}
-		else
-			estlen = SYSLOG_BUFSIZE;
-
-		if (!tracechain || (SYSLOG_BUFSIZE - estlen) <= 0)
-		{
-			/* buffer may be full */
-			if (tracechain && rcptp->chain != NULL)
-				strcat(addrbuf+addrlen, ",(more...)");
-			log(LOG_INFO, "(%d+%d+%d+%d/%d) relay=%s to=%s proto=%s delay=%d code=%d (%s)",
-				sti.nsent, sti.ndeferred, sti.nnsfailed,
-				sti.nsmtpfailed, sti.nrcpt, hostname,
-				addrbuf, proto, delay, rcode,
-				(response==NULL)?"?":response);
-			addrlen = 0;
-			addrbuf[0] = '\0';
-		}
 		if (tracechain == 0)
 			break;
 		rcptp = rcptp->chain;
 	}
+	if (addrlen > 0)
+	{
+		log(LOG_INFO, "(%d+%d+%d+%d/%d) relay=%s to=%s proto=%s delay=%d code=%d (%s)",
+			sti.nsent, sti.ndeferred, sti.nnsfailed,
+			sti.nsmtpfailed, sti.nrcpt, hostname,
+			addrbuf, proto, delay, rcode,
+			(response==NULL)?"?":response);
+	}
 	/* record the time if the most of addresses are processed */
 	if ((sti.time_almostfinish == 0)
 	 && (sti.nsent + sti.ndeferred + sti.nnsfailed + sti.nsmtpfailed
@@ -1174,7 +1249,16 @@ struct recipient *rcptp;
 			if (rcptp2->stat == RCPT_NOOP
 			 || rcptp2->stat == RCPT_RETRY)
 			{
-				if (rcpts + 1 >= cnf.rcpts_trans)
+#if 0
+				if (rcpts + 1 > cnf.rcpts_trans)
+#else
+				if (cnf.debug & DEBUG_TRANS)
+				log(LOG_DEBUG, "check count of rcpts (domain:%s, max rcpts:%d, count:%d)",
+					rcptp2->domain->name,
+					rcptp2->domain->rcpts,
+					rcpts);
+				if (rcpts + 1 > rcptp2->domain->rcpts)
+#endif
 				{
 					/* limited by RFC821 */
 					connwork[i].reuse = 1;
@@ -1261,7 +1345,17 @@ int next;	/* true: forward MX for next trial */
 		if (next)
 		{
 			/* forward MX pointer for next trial */
-			domp->curmx = domp->curmx->next;
+			if (domp->curmx->opt & MAPOPT_TRYALLADDR
+			 && domp->curmx->host->current != NULL)
+			{
+				if (domp->curmx->host->current->next == NULL)
+					domp->curmx = domp->curmx->next;
+				else
+					domp->curmx->host->current
+					= domp->curmx->host->current->next;
+			}
+			else
+				domp->curmx = domp->curmx->next;
 #if 0
 			while (domp->curmx != NULL)
 			{
@@ -1292,9 +1386,10 @@ int next;	/* true: forward MX for next trial */
 	{
 		if (mxp->domain == NULL)
 			continue;
-		if (mxp->domain->curmx == NULL || mxp->domain->rcpt_top == NULL)
+		if (mxp->domain->rcpt_top == NULL)
 			continue;
-		if (mxp->domain->curmx->host != conn->host)
+		if (mxp->domain->curmx != NULL
+		 && mxp->domain->curmx->host != conn->host)
 			continue;
 		trim_rcpts(mxp->domain);
 	}
@@ -1454,6 +1549,10 @@ struct connection *conn;
 		switch (conn->host->current->domain)
 		{
 		  case AF_INET:
+#ifdef INET6
+			if (!(cnf.inetdom & SMTP_V4))
+				goto skip_proto;
+#endif
 			rsin.in.sin_family = AF_INET;
 			bcopy(conn->host->current->address, &rsin.in.sin_addr,
 				conn->host->current->len);
@@ -1476,6 +1575,9 @@ struct connection *conn;
 			break;
 #ifdef INET6
 		  case AF_INET6:
+			if (!(cnf.inetdom & SMTP_V6))
+				goto skip_proto;
+
 			rsin.in6.sin6_family = AF_INET6;
 			bcopy(conn->host->current->address, &rsin.in6.sin6_addr,
 				conn->host->current->len);
@@ -1497,13 +1599,14 @@ struct connection *conn;
 
 				inet_ntop(AF_INET6, &rsin.in6.sin6_addr,
 					buf, sizeof(buf));
-				log(LOG_DEBUG, "%d/-: destination=IPv6:%s/%d",
+				log(LOG_DEBUG, "%d/-: destination=%s/%d",
 					conn->seq, buf, cnf.dst_port);
 			}
 			found = 1;
 			break;
 #endif
 		  default:
+  skip_proto:
 			/* protocol is not supported */
 			conn->host->current->stat = EX_TEMPFAIL;
 			conn->host->current = conn->host->current->next;
@@ -1650,7 +1753,7 @@ struct connection *conn;
 
 					inet_ntop(AF_INET6, &lsin.in6.sin6_addr,
 						buf, sizeof(buf));
-					log(LOG_INFO, "using IPv6:%s as source IP address", buf);
+					log(LOG_INFO, "using %s as source IP address", buf);
 					break;
 #endif
 				}
@@ -1835,7 +1938,8 @@ struct connection *conn;
 		log(LOG_INFO, "%d/%d: getpeername failed for %s: %s",
 			conn->seq, conn->socket, conn->host->name,
 			smtpstrerror(errno));
-		conn->host->current->stat = EX_TEMPFAIL;
+		if (conn->host->current != NULL) /* XXX for safe access */
+			conn->host->current->stat = EX_TEMPFAIL;
 		smtp_close(conn->socket);
 		conn->host->state = STAT_CLOSED;
 		switch (errno)
@@ -1863,7 +1967,8 @@ struct connection *conn;
 			errmsg = smtpstrerror(errno);
 		}
 		markstatus(conn->rcpt, SMTP_TEMPFAIL(52), errmsg, 1);
-		conn->host->current = conn->host->current->next;
+		if (conn->host->current != NULL) /* XXX for safe access */
+			conn->host->current = conn->host->current->next;
 		if (Terminate)
 			return -1;
 		return smtp_connect(conn);
@@ -1880,8 +1985,11 @@ struct connection *conn;
 	if (conn->inbuf == NULL || conn->outbuf == NULL)
 	{
 		log(LOG_NOTICE, "out of memory (SMTP buffer allocation)");
-		conn->host->current->stat = EX_TEMPFAIL;
-		conn->host->current = conn->host->current->next;
+		if (conn->host->current != NULL) /* XXX for safe access */
+		{
+			conn->host->current->stat = EX_TEMPFAIL;
+			conn->host->current = conn->host->current->next;
+		}
 		smtp_close(conn->socket);
 		if (conn->host->current == NULL)
 			conn->host->state = STAT_FAIL;
@@ -2348,17 +2456,20 @@ struct connection *conn;
 		log(LOG_INFO, "%d/%d: failed for %s: %s (%s)",
 			conn->seq, conn->socket, conn->host->name,
 			smtpstrerror(errno), STATE(savestate));
-		conn->host->current->stat = EX_TEMPFAIL;
+		if (conn->host->current != NULL) /* XXX for safe access */
+			conn->host->current->stat = EX_TEMPFAIL;
 		smtp_close(conn->socket);
 		conn->host->state = STAT_CLOSED;
 		if (errmsg != NULL)
 		{
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(52), errmsg, 1);
-			if (conn->host->current->next == NULL
+			if (conn->host->current != NULL /* XXX for safe access */
+			 && conn->host->current->next == NULL
 			 && conn->host->response == NULL)
 				conn->host->response = (char *)errmsg;
 		}
-		conn->host->current = conn->host->current->next;
+		if (conn->host->current != NULL) /* XXX for safe access */
+			conn->host->current = conn->host->current->next;
 		if (Terminate)
 			return -1;
 		return smtp_connect(conn);
@@ -2600,9 +2711,9 @@ struct connection *conn;
 	 && conn->rcpt->chain == NULL)
 	{
 		/* only one recipient */
-		char su[MAXLINE], sd[MAXLINE];
-		char ru[MAXLINE], rd[MAXLINE];
-		char na[MAXLINE*2];
+		char su[MAXLINE], sd[MAXLINE]; /* sender user/domain part */
+		char ru[MAXLINE], rd[MAXLINE]; /* recipient user/domain part */
+		char na[MAXLINE*2]; /* new address */
 		char *p, *q;
 
 		if (strlen(env.sender) > MAXLINE)
@@ -2614,6 +2725,10 @@ struct connection *conn;
 		while (*p != '\0' && *p != '@' && *p != '>')
 			*q++ = *p++;
 		*q = '\0';
+		if (su[0] == '\0')
+		{
+			goto normal_sender;
+		}
 		if (*p++ == '@')
 		{
 			q = sd;
@@ -2654,6 +2769,7 @@ struct connection *conn;
 	}
 	else
 	{
+  normal_sender:
 		if (len < (ll = strlen(env.sender)))
 			goto mail_err;
 		strcat(cmdbuf, env.sender);
@@ -3288,7 +3404,12 @@ struct connection *conn;
 		return -1;	/* to close transaction */
 	}
 
-	conn->host->state = STAT_FAIL;
+	if (rcode == 550 || rcode == 552) {
+		/* for bogus mail gateway to mobile phone.... */
+		conn->host->state = STAT_CLOSED;
+	} else {
+		conn->host->state = STAT_FAIL;
+	}
 	conn->host->stat = rcode;
 	conn->host->response = rstr;
 	smtp_quit_s(conn);
@@ -3454,15 +3575,23 @@ struct connection *conn;
 	switch (conn->host->state)
 	{
 	  case STAT_CONNECTING:
-		conn->host->current->stat = EX_TEMPFAIL;
 		if (conn->host->current == NULL)
+		{
+			conn->host->current->stat = EX_TEMPFAIL;
 			conn->host->state = STAT_FAIL;
+		}
 		else
 			conn->host->state = STAT_CLOSED;
-		if (Terminate || conn->host->current->next == NULL)
+		if (Terminate
+		 || conn->host->current == NULL /* XXX for safe access */
+		 || conn->host->current->next == NULL)
 			markstatus(conn->rcpt, SMTP_TEMPFAIL(52),
 				newstr(buf), 1);
-		conn->host->current = conn->host->current->next;
+		else
+			log(LOG_INFO, "Trying next address of %s",
+			    conn->host->name);
+		if (conn->host->current != NULL) /* XXX for safe access */
+			conn->host->current = conn->host->current->next;
 		if (Terminate)
 			return -1;
 		return smtp_connect(conn);
@@ -3479,11 +3608,23 @@ struct connection *conn;
 	  case STAT_QUITSENT:
 	  case STAT_QUITREOPEN:
 	  case STAT_FAILQUIT:
-		conn->host->state = STAT_FAIL;
-		conn->host->stat = EX_TEMPFAIL;
-		if (conn->rcpt != NULL)
-			markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
-				newstr(buf), 1);
+		if (conn->rcpt != NULL && conn->rcpt->domain->curmx != NULL
+		 && conn->rcpt->domain->curmx->opt & MAPOPT_TRYALLADDR
+		 && conn->host->current->next != NULL)
+		{
+			conn->host->state = STAT_CLOSED;
+			conn->host->current->stat = EX_TEMPFAIL;
+			log(LOG_INFO, "Trying next address of %s",
+			    conn->host->name);
+		}
+		else
+		{
+			conn->host->state = STAT_FAIL;
+			conn->host->stat = EX_TEMPFAIL;
+			if (conn->rcpt != NULL)
+				markstatus(conn->rcpt, SMTP_TEMPFAIL(51),
+					newstr(buf), 1);
+		}
 		smtp_close(conn->socket);
 		return -1;
 
