@@ -1,7 +1,7 @@
 /*
- * Copyright (c) 1997,1998 Motonori Nakamura <motonori@econ.kyoto-u.ac.jp>
- * Copyright (c) 1997,1998 WIDE Project
- * Copyright (c) 1997,1998 Kyoto University
+ * Copyright (c) 1997-2002 Motonori Nakamura <motonori@media.kyoto-u.ac.jp>
+ * Copyright (c) 1997-2002 WIDE Project
+ * Copyright (c) 1997-2002 Kyoto University
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -35,7 +35,7 @@
  */
 
 #ifndef lint
-static char *_id_ = "$Id: util.c,v 1.13 1999/01/22 08:32:21 motonori Exp $";
+static char *_id_ = "$Id: util.c,v 1.19 2002/01/20 14:19:39 motonori Exp $";
 #endif
 
 # include "common.h"
@@ -218,6 +218,148 @@ struct domain *newdomp;
 }
 #endif
 
+static struct domain_rcpts *hash_domain_rcpts[HASHSIZE];
+
+int
+hash_domain_rcpts_lookup(name)
+char *name;
+{
+	struct domain_rcpts *dom_r_p;
+	int idx = hash_func(name);
+
+	if (cnf.debug & DEBUG_HASH)
+	log(LOG_DEBUG, "look up domain(%s) index(%d)", name, idx);
+	dom_r_p = hash_domain_rcpts[idx];
+	while (dom_r_p != NULL)
+	{
+		if (strcasecmp(dom_r_p->name, name) == 0)
+		{
+			if (cnf.debug & DEBUG_HASH)
+			log(LOG_DEBUG, "hit domain(%s) rcpts(%d) in hash table",
+				name, dom_r_p->rcpts);
+			return dom_r_p->rcpts;
+		}
+		dom_r_p = dom_r_p->hash;
+	}
+	if (cnf.debug & DEBUG_HASH)
+	log(LOG_DEBUG, "domain(%s) not found in hash table, return %d",
+		name, cnf.rcpts_trans);
+	return cnf.rcpts_trans;
+}
+
+void
+hash_domain_rcpts_enter(name, newdom_r_p)
+char *name;
+struct domain_rcpts *newdom_r_p;
+{
+	int idx = hash_func(name);
+
+	if (cnf.debug & DEBUG_HASH)
+	log(LOG_DEBUG, "enter domain(%s) index(%d) rcpts(%d) into hash table",
+		name, idx, newdom_r_p->rcpts);
+
+	newdom_r_p->hash = hash_domain_rcpts[idx];
+	hash_domain_rcpts[idx] = newdom_r_p;
+}
+
+int
+isnumeric(p)
+char *p;
+{
+	while(*p != '\0')
+	{
+		if(!isdigit(*p))
+			return 0;
+		p++;
+	}
+	return 1;
+}
+
+#define DEFAULTRCPTS  (char *)"DEFAULT"
+
+int
+rcpts_def_load()
+{
+	FILE *fp;
+	char buf[MAXLINE], *p;
+	int n;
+	struct domain_rcpts *dom_r_p;
+	char *strp, *argp;
+
+	/* open cnf.rcpts_def */
+	fp = fopen(cnf.rcpts_def, "r");
+	if (fp == NULL)
+	{
+		log(LOG_NOTICE, "file %s not found", cnf.rcpts_def);
+		return -1;
+	}
+	/* read cnf.rcpts_def */
+	while (fgets(buf, sizeof(buf), fp) != NULL)
+	{
+		n = strlen(buf);
+		if (n > 0)
+			n--;
+		if (buf[n] == '\n')
+			buf[n--] = '\0';
+		if ((p = strchr(buf, '#')) != NULL)
+		{
+			*p-- = '\0';
+			while (buf < p && isspace(*p))
+				*p-- = '\0';
+		}
+		if (buf[0] == '\0' /* || buf[0] == '#' */)
+			continue;
+
+		/* get domain name and # of recipients */
+		n = strlen(buf);
+		p = buf;
+		while(p < buf+n && ! isspace(*p))
+			*p++;
+		if(p == buf+n)
+			continue;
+		else
+			*p = '\0';
+		strp = buf;		/* domain name */
+		*p++;
+		while(p < buf+n && isspace(*p))
+			*p++;
+		if(p == buf+n)
+			continue;
+		argp = p;		/* # of recipients */
+		while(p < buf+n && ! isspace(*p))
+			*p++;
+		if(p < buf+n)
+			*p = '\0';
+
+		/* set domain name and # of recipients */
+		if(strp != NULL && argp != NULL && isnumeric(argp))
+		{
+			dom_r_p = (struct domain_rcpts*) MALLOC(sizeof(struct domain_rcpts));
+			if (dom_r_p == NULL)
+			{
+				log(LOG_NOTICE, "out of memory (domain_rcpts)");
+				fclose(fp);
+				return -1;
+			}
+			bzero(dom_r_p, sizeof(struct domain_rcpts));
+			dom_r_p->name  = newstr(strp);
+			if(dom_r_p->name == NULL)
+			{
+				log(LOG_NOTICE, "out of memory (dom_r_p->name)");
+				fclose(fp);
+				return -1;
+			}
+			dom_r_p->rcpts = atoi(argp);
+			hash_domain_rcpts_enter(dom_r_p->name, dom_r_p);
+		}
+	}
+	fclose(fp);
+	cnf.rcpts_trans = hash_domain_rcpts_lookup(DEFAULTRCPTS);
+	if (cnf.debug & DEBUG_TRANS)
+	log(LOG_DEBUG, "set cnf.rcpts_trans = %d", cnf.rcpts_trans);
+	return 0;
+}
+
 static struct host *hash_host[HASHSIZE];
 
 struct host *
@@ -326,17 +468,18 @@ host_map_load()
 	while (fgets(buf, sizeof(buf), fp) != NULL)
 	{
 		n = strlen(buf);
-		if (n > 0)
-			n--;
-		if (buf[n] == '\n')
-			buf[n--] = '\0';
+		if (n >= sizeof(buf))
+			buf[--n] = '\0';
+		if (buf[n-1] == '\n')
+			buf[--n] = '\0';
 		if ((p = strchr(buf, '#')) != NULL)
 		{
 			*p-- = '\0';
 			while (buf < p && isspace(*p))
 				*p-- = '\0';
+			n = strlen(buf);
 		}
-		if (buf[0] == '\0' || buf[0] == '#')
+		if (buf[0] == '\0' /* || buf[0] == '#' */)
 			continue;
 		/* count number of columns */
 		p = buf;
@@ -386,6 +529,7 @@ host_map_load()
 		strcpy(strp, q);
 		strp += strlen(strp) + 1;
 
+		/* skip separator spaces */
 		while (*p != '\0')
 		{
 			if (!isspace(*p))
@@ -393,6 +537,7 @@ host_map_load()
 			p++;
 		}
 
+		/* parse "arg" portion */
 		while ((q = strchr(p, ':')) != NULL)
 		{
 			if (q == p)
@@ -406,6 +551,7 @@ host_map_load()
 			strp += strlen(strp) + 1;
 			p = q;
 		}
+		/* final "arg" portion */
 		if (*p != '\0')
 		{
 			strcpy(strp, p);
@@ -446,7 +592,8 @@ char **
 host_map_lookup(domain)
 char *domain;
 {
-	static char *default_map[] = {"MX", NULL}, *root = ".";
+	static char *default_map[] = {"MX", NULL};
+	static char *root = ".";
 	struct host_map *mapp;
 	char *p;
 
@@ -464,7 +611,7 @@ char *domain;
 			char **mp = mapp->arg;
 			int found = 0;
 
-			/* is there a aliasing? */
+			/* is there an aliasing? */
 			if (*p == '.')
 			{
 
